package java.time.chrono;

import java.time.Instant;
import java.time.format.DateTimeFormatter;
import java.time.temporal.TemporalQuery;
import java.time.temporal.TemporalAmount;
import java.time.temporal.TemporalAdjuster;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalUnit;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.LocalTime;
import java.time.temporal.UnsupportedTemporalTypeException;
import java.time.temporal.ChronoField;
import java.time.temporal.ValueRange;
import java.time.temporal.TemporalField;
import java.time.DateTimeException;
import java.time.temporal.TemporalQueries;
import java.util.Objects;
import java.time.temporal.TemporalAccessor;
import java.util.Comparator;
import java.time.temporal.Temporal;

public interface ChronoZonedDateTime<D extends ChronoLocalDate> extends Temporal, Comparable<ChronoZonedDateTime<?>>
{
    public static Comparator<ChronoZonedDateTime<?>> timeLineOrder() {
        return AbstractChronology.INSTANT_ORDER;
    }
    
    public static ChronoZonedDateTime<?> from(TemporalAccessor obj) {
        if ((obj instanceof ChronoZonedDateTime)) {
            return (ChronoZonedDateTime)obj;
        }
        Objects.requireNonNull(obj, "temporal");
        Chronology chronology = (Chronology)obj.query(TemporalQueries.chronology());
        if (chronology == null) {
            throw new DateTimeException("Unable to obtain ChronoZonedDateTime from TemporalAccessor: " + obj.getClass());
        }
        return chronology.zonedDateTime(obj);
    }
    
    default public ValueRange range(TemporalField field) {
        if ((field instanceof ChronoField)) {
            if (field == ChronoField.INSTANT_SECONDS || field == ChronoField.OFFSET_SECONDS) {
                return field.range();
            }
            return this.toLocalDateTime().range(field);
        }
        else {
            return field.rangeRefinedBy(this);
        }
    }
    
    default public int get(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            switch ((ChronoField)temporalField) {
                case INSTANT_SECONDS: {
                    throw new UnsupportedTemporalTypeException("Invalid field 'InstantSeconds' for get() method, use getLong() instead");
                }
                case OFFSET_SECONDS: {
                    return this.getOffset().getTotalSeconds();
                }
                default: {
                    return this.toLocalDateTime().get(temporalField);
                }
            }
        }
        else {
            return super.get(temporalField);
        }
    }
    
    default public long getLong(TemporalField temporalField) {
        if ((temporalField instanceof ChronoField)) {
            switch ((ChronoField)temporalField) {
                case INSTANT_SECONDS: {
                    return this.toEpochSecond();
                }
                case OFFSET_SECONDS: {
                    return (long)this.getOffset().getTotalSeconds();
                }
                default: {
                    return this.toLocalDateTime().getLong(temporalField);
                }
            }
        }
        else {
            return temporalField.getFrom(this);
        }
    }
    
    default public D toLocalDate() {
        return this.toLocalDateTime().toLocalDate();
    }
    
    default public LocalTime toLocalTime() {
        return this.toLocalDateTime().toLocalTime();
    }
    
    public abstract ChronoLocalDateTime<D> toLocalDateTime();
    
    default public Chronology getChronology() {
        return this.toLocalDate().getChronology();
    }
    
    public abstract ZoneOffset getOffset();
    
    public abstract ZoneId getZone();
    
    public abstract ChronoZonedDateTime<D> withEarlierOffsetAtOverlap();
    
    public abstract ChronoZonedDateTime<D> withLaterOffsetAtOverlap();
    
    public abstract ChronoZonedDateTime<D> withZoneSameLocal(ZoneId p0);
    
    public abstract ChronoZonedDateTime<D> withZoneSameInstant(ZoneId p0);
    
    public abstract boolean isSupported(TemporalField p0);
    
    default public boolean isSupported(TemporalUnit temporalUnit) {
        if ((temporalUnit instanceof ChronoUnit)) {
            return temporalUnit != ChronoUnit.FOREVER;
        }
        return temporalUnit != null && temporalUnit.isSupportedBy(this);
    }
    
    default public ChronoZonedDateTime<D> with(TemporalAdjuster adjuster) {
        return ChronoZonedDateTimeImpl.ensureValid(this.getChronology(), super.with(adjuster));
    }
    
    public abstract ChronoZonedDateTime<D> with(TemporalField p0, long p1);
    
    default public ChronoZonedDateTime<D> plus(TemporalAmount amount) {
        return ChronoZonedDateTimeImpl.ensureValid(this.getChronology(), super.plus(amount));
    }
    
    public abstract ChronoZonedDateTime<D> plus(long p0, TemporalUnit p1);
    
    default public ChronoZonedDateTime<D> minus(TemporalAmount amount) {
        return ChronoZonedDateTimeImpl.ensureValid(this.getChronology(), super.minus(amount));
    }
    
    default public ChronoZonedDateTime<D> minus(long amountToSubtract, TemporalUnit unit) {
        return ChronoZonedDateTimeImpl.ensureValid(this.getChronology(), super.minus(amountToSubtract, unit));
    }
    
    default public <R> R query(TemporalQuery<R> temporalQuery) {
        if (temporalQuery == TemporalQueries.zone() || temporalQuery == TemporalQueries.zoneId()) {
            return this.getZone();
        }
        if (temporalQuery == TemporalQueries.offset()) {
            return this.getOffset();
        }
        if (temporalQuery == TemporalQueries.localTime()) {
            return this.toLocalTime();
        }
        if (temporalQuery == TemporalQueries.chronology()) {
            return this.getChronology();
        }
        if (temporalQuery == TemporalQueries.precision()) {
            return ChronoUnit.NANOS;
        }
        return temporalQuery.queryFrom(this);
    }
    
    default public String format(DateTimeFormatter obj) {
        Objects.requireNonNull(obj, "formatter");
        return obj.format(this);
    }
    
    default public Instant toInstant() {
        return Instant.ofEpochSecond(this.toEpochSecond(), (long)this.toLocalTime().getNano());
    }
    
    default public long toEpochSecond() {
        return this.toLocalDate().toEpochDay() * 86400L + (long)this.toLocalTime().toSecondOfDay() - (long)this.getOffset().getTotalSeconds();
    }
    
    default public int compareTo(ChronoZonedDateTime<?> chronoZonedDateTime) {
        int n = Long.compare(this.toEpochSecond(), chronoZonedDateTime.toEpochSecond());
        if (n == 0) {
            n = this.toLocalTime().getNano() - chronoZonedDateTime.toLocalTime().getNano();
            if (n == 0) {
                n = this.toLocalDateTime().compareTo(chronoZonedDateTime.toLocalDateTime());
                if (n == 0) {
                    n = this.getZone().getId().compareTo(chronoZonedDateTime.getZone().getId());
                    if (n == 0) {
                        n = this.getChronology().compareTo(chronoZonedDateTime.getChronology());
                    }
                }
            }
        }
        return n;
    }
    
    default public boolean isBefore(ChronoZonedDateTime<?> chronoZonedDateTime) {
        long epochSecond = this.toEpochSecond();
        long epochSecond2 = chronoZonedDateTime.toEpochSecond();
        return epochSecond < epochSecond2 || (epochSecond == epochSecond2 && this.toLocalTime().getNano() < chronoZonedDateTime.toLocalTime().getNano());
    }
    
    default public boolean isAfter(ChronoZonedDateTime<?> chronoZonedDateTime) {
        long epochSecond = this.toEpochSecond();
        long epochSecond2 = chronoZonedDateTime.toEpochSecond();
        return epochSecond > epochSecond2 || (epochSecond == epochSecond2 && this.toLocalTime().getNano() > chronoZonedDateTime.toLocalTime().getNano());
    }
    
    default public boolean isEqual(ChronoZonedDateTime<?> chronoZonedDateTime) {
        return this.toEpochSecond() == chronoZonedDateTime.toEpochSecond() && this.toLocalTime().getNano() == chronoZonedDateTime.toLocalTime().getNano();
    }
    
    public abstract boolean equals(Object p0);
    
    public abstract int hashCode();
    
    public abstract String toString();
}
